@@ -66,6 +66,20 @@ private:
 	}
 };
 
+template <class T1, class T2>
+bool
+operator==(const pair<T1, T2> &lhs, const pair<T1, T2> &rhs)
+{
+	return lhs.first == rhs.first && lhs.second == rhs.second;
+}
+
+template <class T1, class T2>
+bool
+operator!=(const pair<T1, T2> &lhs, const pair<T1, T2> &rhs)
+{
+	return !(lhs == rhs);
+}
+
 } /* namespace detail */
 
 } /* namespace pmem */
