@@ -1,16 +1,12 @@
 // SPDX-License-Identifier: BSD-3-Clause
-/* Copyright 2020, Intel Corporation */
-
-#include "radix.hpp"
+/* Copyright 2020-2021, Intel Corporation */
 
 #include <algorithm>
-#include <random>
+#include <sstream>
 #include <unordered_map>
 #include <unordered_set>
 
-#include <sstream>
-
-static std::mt19937_64 generator;
+#include "radix.hpp"
 
 void
 test_iterators(nvobj::pool<root> &pop)
@@ -18,14 +14,14 @@ test_iterators(nvobj::pool<root> &pop)
 	auto r = pop.root();
 
 	nvobj::transaction::run(pop, [&] {
-		r->radix_int = nvobj::make_persistent<container_int>();
+		r->radix_int = nvobj::make_persistent<cntr_int>();
 		r->radix_int->try_emplace("", 0U);
 		r->radix_int->try_emplace("ab", 1U);
 		r->radix_int->try_emplace("ba", 2U);
 		r->radix_int->try_emplace("a", 3U);
 		r->radix_int->try_emplace("b", 4U);
 
-		r->radix_str = nvobj::make_persistent<container_string>();
+		r->radix_str = nvobj::make_persistent<cntr_string>();
 		r->radix_str->try_emplace("", "");
 
 		r->radix_str->try_emplace(" ", "ab");
@@ -36,9 +32,19 @@ test_iterators(nvobj::pool<root> &pop)
 		r->radix_str->try_emplace("a", "a");
 		r->radix_str->try_emplace("b", "b");
 	});
+	/* Checks that inline_string copy constructor throws pmem::pool_error */
+	try {
+		auto k = r->radix_str->find("ab")->key();
+		(void)k;
+		ASSERT_UNREACHABLE;
+	} catch (pmem::pool_error &) {
+	} catch (...) {
+		ASSERT_UNREACHABLE;
+	}
 
+	/* Test int keys. */
 	auto it = r->radix_int->find("a");
-	UT_ASSERT(nvobj::string_view(it->key()).compare(std::string("a")) == 0);
+	UT_ASSERTeq(nvobj::string_view(it->key()).compare(std::string("a")), 0);
 	UT_ASSERTeq(it->value(), 3);
 
 	++it;
@@ -47,7 +53,7 @@ test_iterators(nvobj::pool<root> &pop)
 	UT_ASSERTeq(it->value(), 1);
 
 	++it;
-	UT_ASSERT(nvobj::string_view(it->key()).compare(std::string("b")) == 0);
+	UT_ASSERTeq(nvobj::string_view(it->key()).compare(std::string("b")), 0);
 	UT_ASSERTeq(it->value(), 4);
 
 	++it;
@@ -55,70 +61,79 @@ test_iterators(nvobj::pool<root> &pop)
 		  0);
 	UT_ASSERTeq(it->value(), 2);
 
+	++it;
+	UT_ASSERT(it == r->radix_int->end());
 	--it;
-	UT_ASSERT(nvobj::string_view(it->key()).compare(std::string("b")) == 0);
+
+	--it;
+	UT_ASSERTeq(nvobj::string_view(it->key()).compare(std::string("b")), 0);
 	UT_ASSERTeq(it->value(), 4);
 
 	--it;
-	UT_ASSERT(nvobj::string_view(it->key()).compare(std::string("ab")) ==
-		  0);
+	UT_ASSERTeq(nvobj::string_view(it->key()).compare(std::string("ab")),
+		    0);
 	UT_ASSERTeq(it->value(), 1);
 
 	--it;
-	UT_ASSERT(nvobj::string_view(it->key()).compare(std::string("a")) == 0);
+	UT_ASSERTeq(nvobj::string_view(it->key()).compare(std::string("a")), 0);
 	UT_ASSERTeq(it->value(), 3);
 
 	--it;
-	UT_ASSERT(nvobj::string_view(it->key()).compare(std::string("")) == 0);
+	UT_ASSERTeq(nvobj::string_view(it->key()).compare(std::string("")), 0);
 	UT_ASSERTeq(it->value(), 0);
+	UT_ASSERT(it == r->radix_int->begin());
 
 	it = r->radix_int->erase(it);
-	UT_ASSERT(nvobj::string_view(it->key()).compare(std::string("a")) == 0);
+	UT_ASSERTeq(nvobj::string_view(it->key()).compare(std::string("a")), 0);
 	UT_ASSERTeq(it->value(), 3);
+	UT_ASSERT(it == r->radix_int->begin());
 
 	(*it).value() = 4;
-	UT_ASSERT(nvobj::string_view(it->key()).compare(std::string("a")) == 0);
+	UT_ASSERTeq(nvobj::string_view(it->key()).compare(std::string("a")), 0);
 	UT_ASSERTeq(it->value(), 4);
 
 	it = r->radix_int->lower_bound("b");
-	UT_ASSERT(nvobj::string_view(it->key()).compare(std::string("b")) == 0);
+	UT_ASSERTeq(nvobj::string_view(it->key()).compare(std::string("b")), 0);
 
 	it = r->radix_int->lower_bound("aa");
-	UT_ASSERT(nvobj::string_view(it->key()).compare(std::string("ab")) ==
-		  0);
+	UT_ASSERTeq(nvobj::string_view(it->key()).compare(std::string("ab")),
+		    0);
 
+	/* Test string keys. */
 	auto it2 = r->radix_str->find("");
 
 	it2 = r->radix_str->lower_bound("aa");
 	auto it3 = it2;
 	it2.assign_val("xx");
 
-	UT_ASSERT(nvobj::string_view(it2->value()).compare("xx") == 0);
-	UT_ASSERT(nvobj::string_view(it3->value()).compare("xx") == 0);
+	UT_ASSERTeq(nvobj::string_view(it2->value()).compare("xx"), 0);
+	UT_ASSERTeq(nvobj::string_view(it3->value()).compare("xx"), 0);
 
 	auto long_string = std::string(1024, 'x');
 	/* The previous assignment should not invalidate the iterator. */
 	it2.assign_val(long_string);
 
-	UT_ASSERT(nvobj::string_view(it2->value()).compare(long_string) == 0);
+	UT_ASSERTeq(nvobj::string_view(it2->value()).compare(long_string), 0);
 
-	UT_ASSERT(nvobj::string_view(r->radix_str->lower_bound("aa")->value())
-			  .compare(long_string) == 0);
+	UT_ASSERTeq(nvobj::string_view(r->radix_str->lower_bound("aa")->value())
+			    .compare(long_string),
+		    0);
 
 	UT_ASSERT(r->radix_str->find("") != r->radix_str->end());
+	UT_ASSERT(r->radix_str->find("") == r->radix_str->begin());
 	UT_ASSERT(r->radix_str->find(" ") != r->radix_str->end());
-	UT_ASSERT(r->radix_str->find(" ") != r->radix_str->end());
+	UT_ASSERT(r->radix_str->find("  ") != r->radix_str->end());
 
 	/* Verify operator<< compiles. */
 	std::stringstream ss;
 	ss << *r->radix_str;
 
 	nvobj::transaction::run(pop, [&] {
-		nvobj::delete_persistent<container_string>(r->radix_str);
-		nvobj::delete_persistent<container_int>(r->radix_int);
+		nvobj::delete_persistent<cntr_string>(r->radix_str);
+		nvobj::delete_persistent<cntr_int>(r->radix_int);
 	});
 
-	UT_ASSERT(OID_IS_NULL(pmemobj_first(pop.handle())));
+	UT_ASSERTeq(num_allocs(pop), 0);
 }
 
 void
@@ -127,8 +142,8 @@ test_ref_stability(nvobj::pool<root> &pop)
 	auto r = pop.root();
 
 	nvobj::transaction::run(pop, [&] {
-		r->radix_str = nvobj::make_persistent<container_string>();
-		r->radix_int = nvobj::make_persistent<container_int>();
+		r->radix_str = nvobj::make_persistent<cntr_string>();
+		r->radix_int = nvobj::make_persistent<cntr_int>();
 	});
 
 	{
@@ -139,13 +154,15 @@ test_ref_stability(nvobj::pool<root> &pop)
 		auto &acxxxz_ref =
 			*r->radix_str->emplace("acxxxz", "acxxxz").first;
 
-		UT_ASSERT(nvobj::string_view(ab_ref.value()).compare("ab") ==
-			  0);
-		UT_ASSERT(nvobj::string_view(a_ref.value()).compare("a") == 0);
-		UT_ASSERT(nvobj::string_view(acxxxy_ref.value())
-				  .compare("acxxxy") == 0);
-		UT_ASSERT(nvobj::string_view(acxxxz_ref.value())
-				  .compare("acxxxz") == 0);
+		UT_ASSERTeq(nvobj::string_view(ab_ref.value()).compare("ab"),
+			    0);
+		UT_ASSERTeq(nvobj::string_view(a_ref.value()).compare("a"), 0);
+		UT_ASSERTeq(nvobj::string_view(acxxxy_ref.value())
+				    .compare("acxxxy"),
+			    0);
+		UT_ASSERTeq(nvobj::string_view(acxxxz_ref.value())
+				    .compare("acxxxz"),
+			    0);
 	}
 
 	{
@@ -154,16 +171,18 @@ test_ref_stability(nvobj::pool<root> &pop)
 		auto &acxxxy_ref = *r->radix_int->emplace("acxxxy", 3U).first;
 		auto &acxxxz_ref = *r->radix_int->emplace("acxxxz", 4U).first;
 
-		UT_ASSERT(ab_ref.value() == 1U);
-		UT_ASSERT(a_ref.value() == 2U);
-		UT_ASSERT(acxxxy_ref.value() == 3U);
-		UT_ASSERT(acxxxz_ref.value() == 4U);
+		UT_ASSERTeq(ab_ref.value(), 1U);
+		UT_ASSERTeq(a_ref.value(), 2U);
+		UT_ASSERTeq(acxxxy_ref.value(), 3U);
+		UT_ASSERTeq(acxxxz_ref.value(), 4U);
 	}
 
 	nvobj::transaction::run(pop, [&] {
-		nvobj::delete_persistent<container_string>(r->radix_str);
-		nvobj::delete_persistent<container_int>(r->radix_int);
+		nvobj::delete_persistent<cntr_string>(r->radix_str);
+		nvobj::delete_persistent<cntr_int>(r->radix_int);
 	});
+
+	UT_ASSERTeq(num_allocs(pop), 0);
 }
 
 /* Tests some corner cases (not covered by libcxx find/bound tests). */
@@ -174,8 +193,7 @@ test_find(nvobj::pool<root> &pop)
 
 	{
 		nvobj::transaction::run(pop, [&] {
-			r->radix_str =
-				nvobj::make_persistent<container_string>();
+			r->radix_str = nvobj::make_persistent<cntr_string>();
 		});
 
 		UT_ASSERT(r->radix_str->lower_bound("") == r->radix_str->end());
@@ -221,6 +239,10 @@ test_find(nvobj::pool<root> &pop)
 			  r->radix_str->end());
 		UT_ASSERT(r->radix_str->upper_bound("acyyy") ==
 			  r->radix_str->end());
+		UT_ASSERT(r->radix_str->lower_bound("acy") ==
+			  r->radix_str->end());
+		UT_ASSERT(r->radix_str->upper_bound("acy") ==
+			  r->radix_str->end());
 
 		/* Look for key which shares some common part with leafs but
 		 * differs on compressed bytes. */
@@ -236,6 +258,10 @@ test_find(nvobj::pool<root> &pop)
 			  r->radix_str->find("acxxxy"));
 		UT_ASSERT(r->radix_str->upper_bound("acxxa") ==
 			  r->radix_str->find("acxxxy"));
+		UT_ASSERT(r->radix_str->lower_bound("acxxyy") ==
+			  r->radix_str->end());
+		UT_ASSERT(r->radix_str->upper_bound("acxxyy") ==
+			  r->radix_str->end());
 		UT_ASSERT(r->radix_str->lower_bound("acaaa") ==
 			  r->radix_str->find("acxxxy"));
 		UT_ASSERT(r->radix_str->upper_bound("acaaa") ==
@@ -251,23 +277,25 @@ test_find(nvobj::pool<root> &pop)
 			  r->radix_str->find("ad"));
 		UT_ASSERT(r->radix_str->upper_bound("acyyy") ==
 			  r->radix_str->find("ad"));
+		UT_ASSERT(r->radix_str->lower_bound("acxxzy") ==
+			  r->radix_str->find("ad"));
+		UT_ASSERT(r->radix_str->upper_bound("acxxzy") ==
+			  r->radix_str->find("ad"));
 		UT_ASSERT(r->radix_str->lower_bound("acxxxzz") ==
 			  r->radix_str->find("ad"));
 		UT_ASSERT(r->radix_str->upper_bound("acxxxzz") ==
 			  r->radix_str->find("ad"));
 
 		nvobj::transaction::run(pop, [&] {
-			nvobj::delete_persistent<container_string>(
-				r->radix_str);
+			nvobj::delete_persistent<cntr_string>(r->radix_str);
 		});
 
-		UT_ASSERT(OID_IS_NULL(pmemobj_first(pop.handle())));
+		UT_ASSERTeq(num_allocs(pop), 0);
 	}
 
 	{
 		nvobj::transaction::run(pop, [&] {
-			r->radix_str =
-				nvobj::make_persistent<container_string>();
+			r->radix_str = nvobj::make_persistent<cntr_string>();
 		});
 
 		r->radix_str->emplace("a", "");
@@ -279,10 +307,24 @@ test_find(nvobj::pool<root> &pop)
 			  r->radix_str->find("bccc"));
 		UT_ASSERT(r->radix_str->upper_bound("baaaba") ==
 			  r->radix_str->find("bccc"));
+		UT_ASSERT(r->radix_str->lower_bound("aaaaaaaaa") ==
+			  r->radix_str->find("bccc"));
+		UT_ASSERT(r->radix_str->upper_bound("aaaaaaaaa") ==
+			  r->radix_str->find("bccc"));
 		UT_ASSERT(r->radix_str->lower_bound("ba") ==
 			  r->radix_str->find("bccc"));
 		UT_ASSERT(r->radix_str->upper_bound("ba") ==
 			  r->radix_str->find("bccc"));
+		UT_ASSERT(r->radix_str->lower_bound("bcca") ==
+			  r->radix_str->find("bccc"));
+		UT_ASSERT(r->radix_str->upper_bound("bcca") ==
+			  r->radix_str->find("bccc"));
+		UT_ASSERT(r->radix_str->upper_bound("bccc") ==
+			  r->radix_str->find("bccca"));
+		UT_ASSERT(r->radix_str->lower_bound("bccd") ==
+			  r->radix_str->end());
+		UT_ASSERT(r->radix_str->upper_bound("bccd") ==
+			  r->radix_str->end());
 		UT_ASSERT(r->radix_str->lower_bound("b") ==
 			  r->radix_str->find("bccc"));
 		UT_ASSERT(r->radix_str->upper_bound("b") ==
@@ -293,19 +335,15 @@ test_find(nvobj::pool<root> &pop)
 			  r->radix_str->end());
 
 		nvobj::transaction::run(pop, [&] {
-			nvobj::delete_persistent<container_string>(
-				r->radix_str);
+			nvobj::delete_persistent<cntr_string>(r->radix_str);
 		});
 
-		UT_ASSERT(OID_IS_NULL(pmemobj_first(pop.handle())));
+		UT_ASSERTeq(num_allocs(pop), 0);
 	}
 
 	{
-		auto r = pop.root();
-
 		nvobj::transaction::run(pop, [&] {
-			r->radix_str =
-				nvobj::make_persistent<container_string>();
+			r->radix_str = nvobj::make_persistent<cntr_string>();
 		});
 
 		r->radix_str->try_emplace("Y", "1");
@@ -319,17 +357,17 @@ test_find(nvobj::pool<root> &pop)
 		auto lb = r->radix_str->lower_bound("");
 		auto ub = r->radix_str->upper_bound("@@@@");
 
-		UT_ASSERT(std::distance(lb, ub) == 0);
+		UT_ASSERTeq(std::distance(lb, ub), 0);
 
 		lb = r->radix_str->lower_bound("");
 		ub = r->radix_str->upper_bound("ZZZZ");
 
-		UT_ASSERT(std::distance(lb, ub) == 7);
+		UT_ASSERTeq(std::distance(lb, ub), 7);
 
 		lb = r->radix_str->lower_bound("");
 		ub = r->radix_str->upper_bound("ZZZZ");
 
-		UT_ASSERT(std::distance(lb, ub) == 7);
+		UT_ASSERTeq(std::distance(lb, ub), 7);
 		r->radix_str->clear();
 
 		r->radix_str->try_emplace("A", "1");
@@ -343,12 +381,12 @@ test_find(nvobj::pool<root> &pop)
 		lb = r->radix_str->lower_bound("");
 		ub = r->radix_str->upper_bound("CCCC");
 
-		UT_ASSERT(std::distance(lb, ub) == 7);
+		UT_ASSERTeq(std::distance(lb, ub), 7);
 
 		lb = r->radix_str->lower_bound("");
 		ub = r->radix_str->upper_bound("ZZZZ");
 
-		UT_ASSERT(std::distance(lb, ub) == 7);
+		UT_ASSERTeq(std::distance(lb, ub), 7);
 		r->radix_str->clear();
 
 		r->radix_str->try_emplace("A", "1");
@@ -362,7 +400,7 @@ test_find(nvobj::pool<root> &pop)
 		lb = r->radix_str->lower_bound("");
 		ub = r->radix_str->upper_bound("BB");
 
-		UT_ASSERT(std::distance(lb, ub) == 3);
+		UT_ASSERTeq(std::distance(lb, ub), 3);
 		r->radix_str->clear();
 
 		r->radix_str->try_emplace("A", "1");
@@ -376,7 +414,7 @@ test_find(nvobj::pool<root> &pop)
 		lb = r->radix_str->lower_bound("");
 		ub = r->radix_str->upper_bound("@@@@");
 
-		UT_ASSERT(std::distance(lb, ub) == 0);
+		UT_ASSERTeq(std::distance(lb, ub), 0);
 
 		char ch[] = {char((15 << 4)), 0};
 		r->radix_str->try_emplace(ch, "8");
@@ -384,7 +422,7 @@ test_find(nvobj::pool<root> &pop)
 		lb = r->radix_str->lower_bound("");
 		ub = r->radix_str->upper_bound("@@@@");
 
-		UT_ASSERT(std::distance(lb, ub) == 0);
+		UT_ASSERTeq(std::distance(lb, ub), 0);
 
 		char last_slot[] = {char((15 << 4) | 15), 0};
 		lb = r->radix_str->lower_bound("");
@@ -393,11 +431,99 @@ test_find(nvobj::pool<root> &pop)
 		UT_ASSERTeq(std::distance(lb, ub), 8);
 
 		nvobj::transaction::run(pop, [&] {
-			nvobj::delete_persistent<container_string>(
-				r->radix_str);
+			nvobj::delete_persistent<cntr_string>(r->radix_str);
 		});
 
-		UT_ASSERT(OID_IS_NULL(pmemobj_first(pop.handle())));
+		UT_ASSERTeq(num_allocs(pop), 0);
+	}
+
+	{
+		nvobj::transaction::run(pop, [&] {
+			r->radix_str = nvobj::make_persistent<cntr_string>();
+		});
+
+		r->radix_str->try_emplace(std::string(1, char(1)), "");
+
+		auto ub = r->radix_str->upper_bound(std::string(1, char(-1)));
+		UT_ASSERT(ub == r->radix_str->end());
+
+		nvobj::transaction::run(pop, [&] {
+			nvobj::delete_persistent<cntr_string>(r->radix_str);
+		});
+
+		UT_ASSERTeq(num_allocs(pop), 0);
+	}
+
+	{
+		nvobj::transaction::run(pop, [&] {
+			r->radix_str = nvobj::make_persistent<cntr_string>();
+		});
+
+		r->radix_str->try_emplace(std::string(1, char(-1)), "");
+
+		auto ub = r->radix_str->upper_bound(std::string(1, char(1)));
+		UT_ASSERT(ub == r->radix_str->begin());
+
+		nvobj::transaction::run(pop, [&] {
+			nvobj::delete_persistent<cntr_string>(r->radix_str);
+		});
+
+		UT_ASSERTeq(num_allocs(pop), 0);
+	}
+
+	/* *_bound when there are multiple lesser elements with common prefix */
+	{
+		nvobj::transaction::run(pop, [&] {
+			r->radix_str = nvobj::make_persistent<cntr_string>();
+		});
+
+		r->radix_str->try_emplace("in1", "");
+		r->radix_str->try_emplace("in2", "");
+		r->radix_str->try_emplace("in3", "");
+		r->radix_str->try_emplace("in4", "");
+
+		auto it = r->radix_str->upper_bound("in6");
+		UT_ASSERT(it == r->radix_str->end());
+
+		it = r->radix_str->lower_bound("in6");
+		UT_ASSERT(it == r->radix_str->end());
+
+		r->radix_str->try_emplace("in5", "");
+		r->radix_str->try_emplace("in6", "");
+		r->radix_str->try_emplace("in7", "");
+
+		it = r->radix_str->upper_bound("in9");
+		UT_ASSERT(it == r->radix_str->end());
+
+		it = r->radix_str->lower_bound("in9");
+		UT_ASSERT(it == r->radix_str->end());
+
+		nvobj::transaction::run(pop, [&] {
+			nvobj::delete_persistent<cntr_string>(r->radix_str);
+		});
+
+		UT_ASSERTeq(num_allocs(pop), 0);
+	}
+
+	/* *_bound when there is a single lesser element */
+	{
+		nvobj::transaction::run(pop, [&] {
+			r->radix_str = nvobj::make_persistent<cntr_string>();
+		});
+
+		r->radix_str->try_emplace("in", "");
+
+		auto it = r->radix_str->upper_bound("inA");
+		UT_ASSERT(it == r->radix_str->end());
+
+		it = r->radix_str->lower_bound("inA");
+		UT_ASSERT(it == r->radix_str->end());
+
+		nvobj::transaction::run(pop, [&] {
+			nvobj::delete_persistent<cntr_string>(r->radix_str);
+		});
+
+		UT_ASSERTeq(num_allocs(pop), 0);
 	}
 }
 
@@ -405,7 +531,7 @@ const auto compressed_path_len = 4;
 const auto num_children = 3;
 
 static void
-generate_compressed_tree(nvobj::persistent_ptr<container_string> ptr,
+generate_compressed_tree(nvobj::persistent_ptr<cntr_string> ptr,
 			 std::string prefix, int level)
 {
 	/* it should be just one digit */
@@ -428,11 +554,11 @@ generate_compressed_tree(nvobj::persistent_ptr<container_string> ptr,
 }
 
 static void
-verify_bounds(nvobj::persistent_ptr<container_string> ptr,
+verify_bounds(nvobj::persistent_ptr<cntr_string> ptr,
 	      const std::vector<std::string> &keys)
 {
 	for (size_t i = 0; i < keys.size() - 1; i++) {
-		/* generate key k for which k < keys[i] && k >= keys[i - 1] */
+		/* generate key k, where: k < keys[i] && k > keys[i - 1] */
 		auto k = keys[i];
 		k[k.size() - 1]--;
 
@@ -448,7 +574,7 @@ verify_bounds(nvobj::persistent_ptr<container_string> ptr,
 }
 
 static void
-verify_bounds_key(nvobj::persistent_ptr<container_string> ptr,
+verify_bounds_key(nvobj::persistent_ptr<cntr_string> ptr,
 		  const std::vector<std::string> &keys, const std::string &key)
 {
 	auto expected = std::lower_bound(keys.begin(), keys.end(), key);
@@ -470,7 +596,7 @@ test_compression(nvobj::pool<root> &pop)
 	auto r = pop.root();
 
 	nvobj::transaction::run(pop, [&] {
-		r->radix_str = nvobj::make_persistent<container_string>();
+		r->radix_str = nvobj::make_persistent<cntr_string>();
 	});
 
 	generate_compressed_tree(r->radix_str, "", num_levels);
@@ -528,10 +654,10 @@ test_compression(nvobj::pool<root> &pop)
 	}
 
 	nvobj::transaction::run(pop, [&] {
-		nvobj::delete_persistent<container_string>(r->radix_str);
+		nvobj::delete_persistent<cntr_string>(r->radix_str);
 	});
 
-	UT_ASSERT(OID_IS_NULL(pmemobj_first(pop.handle())));
+	UT_ASSERTeq(num_allocs(pop), 0);
 }
 
 /* Tests some corner cases (not covered by libcxx erase tests). */
@@ -540,54 +666,129 @@ test_erase(nvobj::pool<root> &pop)
 {
 	auto r = pop.root();
 
-	nvobj::transaction::run(pop, [&] {
-		r->radix_str = nvobj::make_persistent<container_string>();
-	});
+	{
+		nvobj::transaction::run(pop, [&] {
+			r->radix_str = nvobj::make_persistent<cntr_string>();
+		});
 
-	std::unordered_set<std::string> set = {"b", "acxxx", "acxxxa",
-					       "acxxx!"};
+		std::unordered_set<std::string> set = {
+			"b", "ab", "acxxa", "acxxx", "acxxxa", "acxxx!"};
 
-	/* Used for testing iterator stability. */
-	std::unordered_map<std::string, typename container_string::iterator>
-		its;
-
-	for (auto &s : set) {
-		auto ret = r->radix_str->emplace(s, s);
-		UT_ASSERT(ret.second);
-		its.emplace(s, ret.first);
-	}
-
-	auto erase = [&](std::string key) {
-		UT_ASSERT(r->radix_str->erase(key) == 1);
-		set.erase(key);
+		/* Used for testing iterator stability. */
+		std::unordered_map<std::string, typename cntr_string::iterator>
+			its;
 
 		for (auto &s : set) {
-			auto it = r->radix_str->find(s);
-			UT_ASSERT(it != r->radix_str->end());
-			UT_ASSERT(nvobj::string_view(it->value()).compare(s) ==
-				  0);
-
-			auto &m_it = its.find(s)->second;
-			UT_ASSERT(nvobj::string_view(m_it->key()).compare(s) ==
-				  0);
-			UT_ASSERT(
-				nvobj::string_view(m_it->value()).compare(s) ==
-				0);
+			auto ret = r->radix_str->emplace(s, s);
+			UT_ASSERT(ret.second);
+			its.emplace(s, ret.first);
 		}
-	};
 
-	erase("acxxxa");
-	erase("acxxx!");
-	erase("acxxx");
-	erase("b");
+		auto erase_one = [&](std::string key) {
+			UT_ASSERTeq(r->radix_str->erase(key), 1);
+			set.erase(key);
 
-	UT_ASSERT(r->radix_str->size() == 0);
+			for (auto &s : set) {
+				auto it = r->radix_str->find(s);
+				UT_ASSERT(it != r->radix_str->end());
+				UT_ASSERTeq(nvobj::string_view(it->value())
+						    .compare(s),
+					    0);
 
-	nvobj::transaction::run(pop, [&] {
-		nvobj::delete_persistent<container_string>(r->radix_str);
-	});
+				auto &m_it = its.find(s)->second;
+				UT_ASSERTeq(nvobj::string_view(m_it->key())
+						    .compare(s),
+					    0);
+				UT_ASSERTeq(nvobj::string_view(m_it->value())
+						    .compare(s),
+					    0);
+			}
+		};
 
-	UT_ASSERT(OID_IS_NULL(pmemobj_first(pop.handle())));
+		UT_ASSERTeq(r->radix_str->erase("ac"), 0);
+		erase_one("acxxxa");
+		UT_ASSERTeq(r->radix_str->erase("acxxxa"), 0);
+		erase_one("acxxa");
+		erase_one("acxxx!");
+		erase_one("ab");
+		erase_one("acxxx");
+		erase_one("b");
+		UT_ASSERTeq(r->radix_str->erase("acxxa"), 0);
+
+		UT_ASSERTeq(r->radix_str->size(), 0);
+
+		nvobj::transaction::run(pop, [&] {
+			nvobj::delete_persistent<cntr_string>(r->radix_str);
+		});
+
+		UT_ASSERTeq(num_allocs(pop), 0);
+	}
+
+	{
+		nvobj::transaction::run(pop, [&] {
+			r->radix_str = nvobj::make_persistent<cntr_string>();
+		});
+
+		std::vector<std::string> elements = {
+			"",	 "acxxx!", "b",	     "ab",
+			"acxxa", "acxxx",  "acxxxa", "x",
+		};
+		size_t value_size = 8;
+		for (auto &e : elements) {
+			std::string value = std::string(value_size, '.');
+			value_size *= 2;
+			auto ret = r->radix_str->emplace(e, value);
+			UT_ASSERT(ret.second);
+		}
+
+		/* sort them, to always remove first element;
+		 * iterate on all left elements from front to back */
+		std::vector<std::string> sorted = elements;
+		std::sort(sorted.begin(), sorted.end());
+
+		for (auto i = elements.size(); i >= 1; --i) {
+			auto it = r->radix_str->erase(
+				r->radix_str->find(sorted.front()));
+			auto it_v = sorted.erase(sorted.begin());
+			UT_ASSERT(it == r->radix_str->begin());
+			while (it != r->radix_str->end()) {
+				UT_ASSERT(nvobj::string_view(it->key()) ==
+					  *it_v);
+				++it;
+				++it_v;
+			}
+		}
+
+		/* set and sort them again, to always remove last element;
+		 * iterate on all left elements from back to front */
+		for (auto &e : elements) {
+			auto ret = r->radix_str->emplace(e, e);
+			UT_ASSERT(ret.second);
+		}
+		sorted = elements;
+		std::sort(sorted.begin(), sorted.end());
+
+		for (auto i = elements.size(); i >= 1; --i) {
+			auto it = r->radix_str->erase(
+				r->radix_str->find(sorted.back()));
+			auto it_v = sorted.erase(--sorted.end());
+			UT_ASSERT(it == r->radix_str->end());
+			while (it != r->radix_str->begin()) {
+				--it;
+				--it_v;
+				UT_ASSERT(nvobj::string_view(it->key()) ==
+					  *it_v);
+			}
+		}
+
+		UT_ASSERTeq(r->radix_str->size(), 0);
+
+		nvobj::transaction::run(pop, [&] {
+			nvobj::delete_persistent<cntr_string>(r->radix_str);
+		});
+
+		UT_ASSERTeq(num_allocs(pop), 0);
+	}
 }
 
 /* This test inserts elements in range [0:2:2 * numeric_limits<uint16_t>::max()]
@@ -599,11 +800,11 @@ test_binary_keys(nvobj::pool<root> &pop)
 	auto kv_f = [](unsigned i) { return i * 2; };
 
 	nvobj::transaction::run(pop, [&] {
-		r->radix_int_int = nvobj::make_persistent<container_int_int>();
+		r->radix_int_int = nvobj::make_persistent<cntr_int_int>();
 	});
 
 	/* Used for testing iterator stability. */
-	std::unordered_map<unsigned, typename container_int_int::iterator> its;
+	std::unordered_map<unsigned, typename cntr_int_int::iterator> its;
 
 	for (unsigned i = 2 * std::numeric_limits<uint16_t>::max(); i > 0;
 	     i -= 2) {
@@ -642,15 +843,15 @@ test_binary_keys(nvobj::pool<root> &pop)
 	}
 
 	nvobj::transaction::run(pop, [&] {
-		nvobj::delete_persistent<container_int_int>(r->radix_int_int);
+		nvobj::delete_persistent<cntr_int_int>(r->radix_int_int);
 	});
 
 	its = {};
 
-	UT_ASSERT(OID_IS_NULL(pmemobj_first(pop.handle())));
+	UT_ASSERTeq(num_allocs(pop), 0);
 
 	nvobj::transaction::run(pop, [&] {
-		r->radix_int_int = nvobj::make_persistent<container_int_int>();
+		r->radix_int_int = nvobj::make_persistent<cntr_int_int>();
 	});
 
 	for (unsigned i = 0; i < 2 * std::numeric_limits<uint16_t>::max();
@@ -684,16 +885,16 @@ test_binary_keys(nvobj::pool<root> &pop)
 	}
 
 	nvobj::transaction::run(pop, [&] {
-		nvobj::delete_persistent<container_int_int>(r->radix_int_int);
+		nvobj::delete_persistent<cntr_int_int>(r->radix_int_int);
 	});
 
-	UT_ASSERT(OID_IS_NULL(pmemobj_first(pop.handle())));
+	UT_ASSERTeq(num_allocs(pop), 0);
 }
 
 void
 test_pre_post_fixes(nvobj::pool<root> &pop)
 {
-	auto num_elements = (1U << 10);
+	size_t num_elements = (1U << 10);
 
 	std::vector<std::string> elements;
 	elements.reserve(num_elements);
@@ -701,14 +902,13 @@ test_pre_post_fixes(nvobj::pool<root> &pop)
 	elements.push_back("0");
 
 	/* Used for testing iterator stability. */
-	std::unordered_map<std::string, typename container_string::iterator>
-		its;
+	std::unordered_map<std::string, typename cntr_string::iterator> its;
 
 	/* This loop creates string so that elements[i] is prefix of
 	 * elements[i + 1] and they differ only by 4 bits:
 	 * '0xA0', '0xA0 0xAB', '0xA0 0xAB 0xC0', '0xA0 0xAB 0xC0 0xCD'
 	 */
-	for (unsigned i = 1; i < num_elements * 2; i++) {
+	for (size_t i = 1; i < num_elements * 2; i++) {
 		if (i % 2 == 0)
 			elements.push_back(
 				elements.back() +
@@ -731,10 +931,10 @@ test_pre_post_fixes(nvobj::pool<root> &pop)
 	auto r = pop.root();
 
 	nvobj::transaction::run(pop, [&] {
-		r->radix_str = nvobj::make_persistent<container_string>();
+		r->radix_str = nvobj::make_persistent<cntr_string>();
 	});
 
-	for (unsigned i = elements.size(); i > 0; i--) {
+	for (size_t i = elements.size(); i > 0; i--) {
 		auto ret =
 			r->radix_str->emplace(elements[i - 1], elements[i - 1]);
 		if (ret.second)
@@ -744,26 +944,28 @@ test_pre_post_fixes(nvobj::pool<root> &pop)
 	verify_bounds(r->radix_str, s_elements);
 
 	UT_ASSERTeq(r->radix_str->size(), num_elements);
-	unsigned i = 0;
+	size_t i = 0;
 	for (auto it = r->radix_str->begin(); it != r->radix_str->end();
 	     ++it, ++i) {
 		UT_ASSERT(nvobj::string_view(it->key()) == s_elements[i]);
 	}
 
 	/* Used for testing iterator stability. */
-	for (unsigned i = num_elements; i > 0; i--) {
+	for (size_t i = num_elements; i > 0; i--) {
 		auto &it = its.find(s_elements[i - 1])->second;
-		UT_ASSERT(nvobj::string_view(it->key()).compare(
-				  s_elements[i - 1]) == 0);
-		UT_ASSERT(nvobj::string_view(it->value())
-				  .compare(s_elements[i - 1]) == 0);
+		UT_ASSERTeq(nvobj::string_view(it->key()).compare(
+				    s_elements[i - 1]),
+			    0);
+		UT_ASSERTeq(nvobj::string_view(it->value())
+				    .compare(s_elements[i - 1]),
+			    0);
 	}
 
 	nvobj::transaction::run(pop, [&] {
-		nvobj::delete_persistent<container_string>(r->radix_str);
+		nvobj::delete_persistent<cntr_string>(r->radix_str);
 	});
 
-	UT_ASSERT(OID_IS_NULL(pmemobj_first(pop.handle())));
+	UT_ASSERTeq(num_allocs(pop), 0);
 }
 
 void
@@ -774,7 +976,7 @@ test_assign_inline_string(nvobj::pool<root> &pop)
 	auto r = pop.root();
 
 	nvobj::transaction::run(pop, [&] {
-		r->radix_str = nvobj::make_persistent<container_string>();
+		r->radix_str = nvobj::make_persistent<cntr_string>();
 		r->radix_str->try_emplace("key", test_value);
 	});
 
@@ -784,67 +986,70 @@ test_assign_inline_string(nvobj::pool<root> &pop)
 		r->radix_str->find("key").assign_val(new_value);
 	}
 
-	UT_ASSERT(nvobj::string_view(r->radix_str->find("key")->value())
-			  .compare(new_value) == 0);
+	UT_ASSERTeq(nvobj::string_view(r->radix_str->find("key")->value())
+			    .compare(new_value),
+		    0);
 
 	nvobj::transaction::run(pop, [&] {
-		nvobj::delete_persistent<container_string>(r->radix_str);
+		nvobj::delete_persistent<cntr_string>(r->radix_str);
 	});
 
-	UT_ASSERT(OID_IS_NULL(pmemobj_first(pop.handle())));
+	UT_ASSERTeq(num_allocs(pop), 0);
 }
 
 void
 test_inline_string_u8t_key(nvobj::pool<root> &pop)
 {
+	const size_t NUM_ITER = 10;
 	auto r = pop.root();
 
 	nvobj::transaction::run(pop, [&] {
 		r->radix_inline_s_u8t =
-			nvobj::make_persistent<container_inline_s_u8t>();
+			nvobj::make_persistent<cntr_inline_s_u8t>();
 	});
 	auto &m = *r->radix_inline_s_u8t;
 
-	UT_ASSERT(m.size() == 0);
+	UT_ASSERTeq(m.size(), 0);
 
-	for (unsigned i = 0; i < 10; i++) {
+	for (size_t i = 0; i < NUM_ITER; i++) {
 		auto key = std::basic_string<uint8_t>(i + 10, 99);
 		auto ret = m.try_emplace(
 			key, std::basic_string<uint8_t>({uint8_t(i)}));
 		UT_ASSERT(ret.second);
-		UT_ASSERT(key.compare(ret.first->key().data()) == 0);
+		UT_ASSERTeq(key.compare(ret.first->key().data()), 0);
 		UT_ASSERT(ret.first->value() ==
 			  std::basic_string<uint8_t>({uint8_t(i)}));
-		UT_ASSERT(m.size() == i + 1);
+		UT_ASSERTeq(m.size(), i + 1);
 	}
 
-	for (unsigned i = 0; i < 10; i++) {
+	for (size_t i = 0; i < NUM_ITER; i++) {
 		auto key = std::basic_string<uint8_t>(i + 10, 99);
 		auto ret = m.insert_or_assign(
 			key, std::basic_string<uint8_t>({uint8_t(i + 1)}));
 		UT_ASSERT(!ret.second);
-		UT_ASSERT(key.compare(ret.first->key().data()) == 0);
+		UT_ASSERTeq(key.compare(ret.first->key().data()), 0);
 		UT_ASSERT(ret.first->value() ==
 			  std::basic_string<uint8_t>({uint8_t(i + 1)}));
-		UT_ASSERT(m.size() == 10);
+		UT_ASSERTeq(m.size(), NUM_ITER);
 	}
 
 	auto key = std::basic_string<uint8_t>(15, 99);
 	auto it = m.find(key);
-	UT_ASSERT(key.compare(it->key().data()) == 0);
+	UT_ASSERTeq(key.compare(it->key().data()), 0);
 	UT_ASSERT(it->value() == std::basic_string<uint8_t>({uint8_t(6)}));
 
 	it = m.erase(it);
-	UT_ASSERT(std::basic_string<uint8_t>(16, 99).compare(
-			  it->key().data()) == 0);
+	UT_ASSERTeq(
+		std::basic_string<uint8_t>(16, 99).compare(it->key().data()),
+		0);
 	UT_ASSERT(it->value() == std::basic_string<uint8_t>({uint8_t(7)}));
 
 	nvobj::transaction::run(pop, [&] {
-		nvobj::delete_persistent<container_inline_s_u8t>(
+		nvobj::delete_persistent<cntr_inline_s_u8t>(
 			r->radix_inline_s_u8t);
 	});
 
-	UT_ASSERT(OID_IS_NULL(pmemobj_first(pop.handle())));
+	UT_ASSERTeq(num_allocs(pop), 0);
 }
 
 void
@@ -854,85 +1059,104 @@ test_inline_string_wchart_key(nvobj::pool<root> &pop)
 
 	nvobj::transaction::run(pop, [&] {
 		r->radix_inline_s_wchart =
-			nvobj::make_persistent<container_inline_s_wchart>();
+			nvobj::make_persistent<cntr_inline_s_wchart>();
 	});
 	auto &m = *r->radix_inline_s_wchart;
 
-	UT_ASSERT(m.size() == 0);
+	UT_ASSERTeq(m.size(), 0);
 
 	auto key1 = std::basic_string<wchar_t>(1, 256);
 	auto key2 = std::basic_string<wchar_t>(1, 0);
 	m.try_emplace(key1, 256U);
 	m.try_emplace(key2, 0U);
-	UT_ASSERT(m.size() == 2);
+	UT_ASSERTeq(m.size(), 2);
 	auto it = m.find(key1);
-	UT_ASSERT(it->value() == 256U);
+	UT_ASSERTeq(it->value(), 256U);
 	it = m.find(key2);
-	UT_ASSERT(it->value() == 0U);
+	UT_ASSERTeq(it->value(), 0U);
 
 	key1 = std::basic_string<wchar_t>(10, 257);
 	key2 = std::basic_string<wchar_t>(10, 1);
 	m.try_emplace(key1, 999U);
 	m.try_emplace(key2, 100U);
-	UT_ASSERT(m.size() == 4);
+	UT_ASSERTeq(m.size(), 4);
 	it = m.find(key1);
-	UT_ASSERT(it->value() == 999U);
+	UT_ASSERTeq(it->value(), 999U);
 	it = m.find(key2);
-	UT_ASSERT(it->value() == 100U);
+	UT_ASSERTeq(it->value(), 100U);
+
+	r->radix_inline_s_wchart->clear();
+	UT_ASSERTeq(m.size(), 0);
 
 	nvobj::transaction::run(pop, [&] {
-		nvobj::delete_persistent<container_inline_s_wchart>(
+		nvobj::delete_persistent<cntr_inline_s_wchart>(
 			r->radix_inline_s_wchart);
 	});
+
+	UT_ASSERTeq(num_allocs(pop), 0);
 }
 
 void
 test_remove_inserted(nvobj::pool<root> &pop)
 {
 	const size_t NUM_ITER = 100;
-
 	auto r = pop.root();
 
 	nvobj::transaction::run(pop, [&] {
-		r->radix_str = nvobj::make_persistent<container_string>();
+		r->radix_str = nvobj::make_persistent<cntr_string>();
 	});
 
 	/* remove element which was just inserted */
 	nvobj::transaction::run(pop, [&] {
 		for (size_t i = 0; i < NUM_ITER; i++) {
-			UT_ASSERT(r->radix_str
-					  ->emplace(std::to_string(i),
-						    std::to_string(i))
-					  .second);
-			UT_ASSERT(r->radix_str->erase(std::to_string(i)));
+			std::string i_str = std::to_string(i);
+			UT_ASSERT(!r->radix_str->erase(i_str));
+			UT_ASSERT(r->radix_str->emplace(i_str, i_str).second);
+			UT_ASSERT(r->radix_str->erase(i_str));
 		}
 	});
 
 	/* insert some initial elements */
 	nvobj::transaction::run(pop, [&] {
-		for (size_t i = 0; i < 5; i++)
-			UT_ASSERT(r->radix_str
-					  ->emplace("init" + std::to_string(i),
-						    std::to_string(i))
+		for (size_t i = 0; i < 5; i++) {
+			std::string i_str = std::to_string(i);
+			UT_ASSERT(r->radix_str->emplace("init" + i_str, i_str)
 					  .second);
+		}
 	});
 
 	/* remove element which was just inserted */
 	nvobj::transaction::run(pop, [&] {
 		for (size_t i = 0; i < NUM_ITER; i++) {
-			UT_ASSERT(r->radix_str
-					  ->emplace(std::to_string(i),
-						    std::to_string(i))
-					  .second);
-			UT_ASSERT(r->radix_str->erase(std::to_string(i)));
+			std::string i_str = std::to_string(i);
+			UT_ASSERT(r->radix_str->emplace(i_str, i_str).second);
+			UT_ASSERT(r->radix_str->erase(i_str));
 		}
 	});
 
+	r->radix_str->clear();
+	UT_ASSERTeq(r->radix_str->size(), 0);
+
 	nvobj::transaction::run(pop, [&] {
-		nvobj::delete_persistent<container_string>(r->radix_str);
+		nvobj::delete_persistent<cntr_string>(r->radix_str);
 	});
 
-	UT_ASSERT(OID_IS_NULL(pmemobj_first(pop.handle())));
+	UT_ASSERTeq(num_allocs(pop), 0);
+}
+
+void
+test_error_handle(nvobj::pool<root> &pop)
+{
+	auto r = pop.root();
+
+	try {
+		r->radix_str = nvobj::make_persistent<cntr_string>();
+	} catch (pmem::transaction_scope_error &tse) {
+	} catch (...) {
+		UT_ASSERT(0);
+	}
+
+	UT_ASSERTeq(num_allocs(pop), 0);
 }
 
 static void
@@ -953,10 +1177,7 @@ test(int argc, char *argv[])
 		UT_FATAL("!pool::create: %s %s", pe.what(), path);
 	}
 
-	std::random_device rd;
-	auto seed = rd();
-	std::cout << "rand seed: " << seed << std::endl;
-	generator = std::mt19937_64(seed);
+	init_random();
 
 	test_ref_stability(pop);
 	test_iterators(pop);
@@ -969,6 +1190,7 @@ test(int argc, char *argv[])
 	test_inline_string_u8t_key(pop);
 	test_inline_string_wchart_key(pop);
 	test_remove_inserted(pop);
+	test_error_handle(pop);
 
 	pop.close();
 }
