@@ -1,39 +1,11 @@
-/*
- * Copyright 2016-2020, Intel Corporation
- *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
- *
- *     * Redistributions of source code must retain the above copyright
- *       notice, this list of conditions and the following disclaimer.
- *
- *     * Redistributions in binary form must reproduce the above copyright
- *       notice, this list of conditions and the following disclaimer in
- *       the documentation and/or other materials provided with the
- *       distribution.
- *
- *     * Neither the name of the copyright holder nor the names of its
- *       contributors may be used to endorse or promote products derived
- *       from this software without specific prior written permission.
- *
- * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
- * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
- * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
- * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
- * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
- * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
- * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
- * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
- * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
- * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
- * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
- */
+// SPDX-License-Identifier: BSD-3-Clause
+/* Copyright 2016-2020, Intel Corporation */
 
 /*
  * obj_cpp_transaction.cpp -- cpp transaction test
  */
 
+#include "transaction.hpp"
 #include "unittest.hpp"
 
 #include <libpmemobj++/make_persistent.hpp>
@@ -44,33 +16,6 @@
 #include <libpmemobj++/pool.hpp>
 #include <libpmemobj++/shared_mutex.hpp>
 
-namespace
-{
-int counter = 0;
-}
-
-/*
- * XXX The Microsoft compiler does not follow the ISO SD-6: SG10 Feature
- * Test Recommendations. "_MSC_VER" is a workaround.
- */
-#if _MSC_VER < 1900
-#ifndef __cpp_lib_uncaught_exceptions
-#define __cpp_lib_uncaught_exceptions 201411
-namespace std
-{
-
-int
-uncaught_exceptions() noexcept
-{
-	return ::counter;
-}
-
-} /* namespace std */
-#endif /* __cpp_lib_uncaught_exceptions */
-#endif /* _MSC_VER */
-
-#include <libpmemobj++/transaction.hpp>
-
 #define LAYOUT "cpp"
 #define POOL_SIZE 2 * PMEMOBJ_MIN_POOL
 
@@ -217,12 +162,10 @@ test_shared_mutex_self_deadlock()
 	 * Starting transaction with already taken shared_lock should fail.
 	 *
 	 * However:
-	 *  - pmemobj prior to 1.5.1 has a bug (see pmem/pmdk#3536) which
-	 *    corrupts mutex state by unlocking it when it shouldn't
 	 *  - shared_mutexes (rwlocks), as implemented by pmemobj, do not detect
 	 *    self-deadlocks on Windows
 	 */
-#if TESTS_LIBPMEMOBJ_VERSION < 0x010501 || defined(_WIN32)
+#ifdef _WIN32
 	return false;
 #else
 	return true;
@@ -275,31 +218,6 @@ test_tx_throw_no_abort(nvobj::pool<root> &pop)
 	UT_ASSERT(rootp->pfoo == nullptr);
 	UT_ASSERT(rootp->parr == nullptr);
 
-	try {
-		nvobj::transaction::run(pop, [&]() {
-			rootp->pfoo = nvobj::make_persistent<foo>();
-			try {
-				nvobj::transaction::run(pop, [&]() {
-					throw std::runtime_error("error");
-				});
-			} catch (std::runtime_error &) {
-				exception_thrown = true;
-			} catch (...) {
-				UT_ASSERT(0);
-			}
-			UT_ASSERT(exception_thrown);
-			exception_thrown = false;
-		});
-	} catch (pmem::transaction_error &) {
-		exception_thrown = true;
-	} catch (...) {
-		UT_ASSERT(0);
-	}
-
-	UT_ASSERT(exception_thrown);
-	UT_ASSERT(rootp->pfoo == nullptr);
-	UT_ASSERT(rootp->parr == nullptr);
-
 	if (test_shared_mutex_self_deadlock()) {
 		exception_thrown = false;
 		rootp->shared_mutex.lock();
@@ -513,37 +431,6 @@ test_tx_throw_no_abort_scope(nvobj::pool<root> &pop)
 	UT_ASSERT(rootp->pfoo == nullptr);
 	UT_ASSERT(rootp->parr == nullptr);
 
-	try {
-		counter = 0;
-		T to(pop);
-		rootp->pfoo = nvobj::make_persistent<foo>();
-		try {
-			T to_nested(pop);
-			counter = 1;
-			throw std::runtime_error("error");
-		} catch (std::runtime_error &) {
-			exception_thrown = true;
-		} catch (...) {
-			UT_ASSERT(0);
-		}
-		counter = 0;
-		UT_ASSERT(exception_thrown);
-		exception_thrown = false;
-	} catch (pmem::transaction_error &) {
-		exception_thrown = true;
-	} catch (...) {
-		UT_ASSERT(0);
-	}
-
-	/* the transaction will be aborted silently */
-	UT_ASSERTeq(nvobj::transaction::error(), ECANCELED);
-	if (std::is_same<T, nvobj::transaction::automatic>::value)
-		UT_ASSERT(exception_thrown);
-	else
-		UT_ASSERT(!exception_thrown);
-	UT_ASSERT(rootp->pfoo == nullptr);
-	UT_ASSERT(rootp->parr == nullptr);
-
 	/* committing non-existent transaction should fail with an exception */
 	exception_thrown = false;
 	try {
@@ -582,6 +469,15 @@ test_tx_no_throw_abort_scope(nvobj::pool<root> &pop)
 {
 	auto rootp = pop.root();
 
+	if (rootp->pfoo) {
+		nvobj::transaction::run(pop, [&]() {
+			nvobj::delete_persistent<foo>(rootp->pfoo);
+			nvobj::delete_persistent<nvobj::p<int>>(rootp->parr);
+			rootp->pfoo = nullptr;
+			rootp->parr = nullptr;
+		});
+	}
+
 	UT_ASSERT(rootp->pfoo == nullptr);
 	UT_ASSERT(rootp->parr == nullptr);
 
@@ -755,34 +651,6 @@ test_tx_automatic_destructor_throw(nvobj::pool<root> &pop)
 	UT_ASSERT(exception_thrown);
 	UT_ASSERT(rootp->pfoo == nullptr);
 	UT_ASSERT(rootp->parr == nullptr);
-
-	try {
-		counter = 0;
-		nvobj::transaction::automatic to(pop);
-		rootp->pfoo = nvobj::make_persistent<foo>();
-		try {
-			nvobj::transaction::automatic to_nested(pop);
-			counter = 1;
-			throw std::runtime_error("error");
-		} catch (std::runtime_error &) {
-			exception_thrown = true;
-			counter = 0;
-		} catch (...) {
-			UT_ASSERT(0);
-		}
-		UT_ASSERT(exception_thrown);
-		exception_thrown = false;
-	} catch (pmem::transaction_error &) {
-		exception_thrown = true;
-	} catch (...) {
-		UT_ASSERT(0);
-	}
-
-	/* the transaction will be aborted silently */
-	UT_ASSERTeq(nvobj::transaction::error(), ECANCELED);
-	UT_ASSERT(exception_thrown);
-	UT_ASSERT(rootp->pfoo == nullptr);
-	UT_ASSERT(rootp->parr == nullptr);
 }
 
 /*
@@ -1117,7 +985,7 @@ test_tx_callback_scope(nvobj::pool<root> &pop, std::function<void()> commit)
 }
 
 void
-test_tx_callback_outiside_tx()
+test_tx_callback_outside_tx()
 {
 	try {
 		nvobj::transaction::register_callback(
@@ -1130,13 +998,29 @@ test_tx_callback_outiside_tx()
 		UT_ASSERT(0);
 	}
 }
+
+void
+test_tx_run_in_wrong_stage(nvobj::pool<root> &pop)
+{
+	nvobj::transaction::run(pop, [&] {
+		nvobj::transaction::register_callback(
+			nvobj::transaction::stage::oncommit, [&] {
+				try {
+					nvobj::transaction::run(
+						pop, [&] { UT_ASSERT(0); });
+					UT_ASSERT(0);
+				} catch (pmem::transaction_scope_error &) {
+				} catch (...) {
+					UT_ASSERT(0);
+				}
+			});
+	});
+}
 }
 
-int
-main(int argc, char *argv[])
+static void
+test(int argc, char *argv[])
 {
-	START();
-
 	if (argc != 2)
 		UT_FATAL("usage: %s file-name", argv[0]);
 
@@ -1172,9 +1056,15 @@ main(int argc, char *argv[])
 	test_tx_callback_scope<nvobj::transaction::manual>(pop, real_commit);
 	test_tx_callback_scope<nvobj::transaction::automatic>(pop, fake_commit);
 
-	test_tx_callback_outiside_tx();
+	test_tx_callback_outside_tx();
+
+	test_tx_run_in_wrong_stage(pop);
 
 	pop.close();
+}
 
-	return 0;
+int
+main(int argc, char *argv[])
+{
+	return run_test([&] { test(argc, argv); });
 }
