@@ -1,6 +1,6 @@
 #!/usr/bin/env bash
 # SPDX-License-Identifier: BSD-3-Clause
-# Copyright 2018-2020, Intel Corporation
+# Copyright 2018-2021, Intel Corporation
 
 #
 # install-pmdk.sh - installs libpmem & libpmemobj
@@ -9,22 +9,23 @@
 
 set -e
 
-PACKAGE_MANAGER=${1}
-
-# common: 1.9.2, 28.10.2020
-PMDK_VERSION="1.9.2"
-
 if [ "${SKIP_PMDK_BUILD}" ]; then
 	echo "Variable 'SKIP_PMDK_BUILD' is set; skipping building PMDK"
 	exit
 fi
 
+PACKAGE_MANAGER=${1}
+
+# master: 1.11.0, 02.07.2021
+PMDK_VERSION="8583fcfd68764ac6779e6f93db89b06971b26704"
+
 git clone https://github.com/pmem/pmdk
 cd pmdk
 git checkout ${PMDK_VERSION}
 
-make -j$(nproc) prefix=/opt/pmdk
-sudo make install -j$(nproc) prefix=/opt/pmdk
+# Don't generate docs, they are reundant for us
+make DOC=n -j$(nproc) prefix=/opt/pmdk
+sudo make DOC=n install -j$(nproc) prefix=/opt/pmdk
 
 # Do not create nor test any packages if PACKAGE_MANAGER is not set
 [[ -z "${PACKAGE_MANAGER}" ]] && exit 0
